import os

from flask import Flask, request
from flask_smorest import Api, Blueprint
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker

from insecure_api.models import Product

db_uri = os.getenv("DB_URI")

assert db_uri is not None, "Need to provide DB_URI environment variable"

session_maker = sessionmaker(bind=create_engine(db_uri))


server = Flask(__name__)
api = Api(server, spec_kwargs={"title": "Products API", "version": "v1", "openapi_version": "3.1"})

bp = Blueprint("products", __name__)


@server.get("/categories")
def list_categories():
    with session_maker() as session:
        categories = session.query(Product.category).distinct()
    return {
        "categories": [category[0] for category in categories]
    }  # "1' OR 1=1 --"


@server.get("/products")
def list_products_without_validating_query_params_and_no_sql_parametrization():
    with session_maker() as session:
        products = session.execute(
            text(f"select * from product where category = '{request.args.get('category', 'shoes')}'")
        )
    return {
        "products": [{
            "id": product.id,
            "name": product.name,
            "price": product.price,
            "category": product.category
        } for product in products]
    }


@server.post("/categories")
def create_category_without_validating_payload():
    """
    schema for the payload:
    Category:
      name:
        type: string
      products:
        type: array
        items:
          $ref: '#/components/schemas/Product'
    Product:
      name:
        type: string
      price:
        type: number
    """

    payload = request.json
    with session_maker() as session:
        pass





# BAD ONES:
# - non-robust validation of query params:
#   > not checking for ony-allowed parameters
#   > not validating their types strictly
#   > not validating their possible values
# - non-validated payloads
# - non-robust validation of payloads
# - payloads with bad design
# - undocumented endpoints
# - too much exposure of IDs through payloads allowing to explore for undocumented resource URLs
# - allow editing resources that I haven't created
# - sql injection
# - unauthenticated endpoints when it should
# - not validating token's signature
# - allow client to set token's signature to None, noNe, none, False, etc.
# - using token signatures
# - no rate limiting



import json
from pathlib import Path

import click
import yaml
from tabulate import tabulate

from .test_runner import TestRunner
from . import __version__
from .api_spec import APISpec


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command()
@click.option('--oas-file', type=click.Path(exists=True))
@click.option('--base-url', type=click.STRING)
def run(oas_file, base_url):
    click.echo(click.style(f"Running Fencer {__version__}", fg="green"))
    click.echo(click.style(f"OpenAPI specification file: {oas_file}", fg="green"))
    click.echo(click.style(f"Base URL: {base_url}", fg="green"))

    if oas_file.endswith(".json"):
        spec = json.loads(Path(oas_file).read_text())
    elif oas_file.endswith(".yaml") or oas_file.endswith(".yml"):
        spec = yaml.safe_load(Path(oas_file).read_text())
    else:
        raise Exception("File format not supported!")
    api_spec = APISpec(base_url=base_url, spec=spec)
    api_spec.load_endpoints()

    click.echo("Creating .fencer/ folder to store failing tests...")
    Path('.fencer').mkdir(exist_ok=True)

    test_runner = TestRunner(api_spec=api_spec)

    injection_message = """
  -------------------------
  Testing injection attacks
  -------------------------"""
    click.echo(injection_message)
    test_runner.run_sql_injection_attacks()

    injection_message = """
  -------------------------
  Testing unauthorized access
  -------------------------"""
    click.echo(injection_message)
    test_runner.run_unauthorized_access_attacks()

    click.echo()

    click.echo(click.style("  SUMMARY", fg="green"))

    click.echo()
    click.echo(click.style("> Number of tests", fg="yellow"))
    click.echo(tabulate({
        "Test Category": [report.category.value for report in test_runner.reports],
        "Number of tests": [report.number_tests for report in test_runner.reports],
        "Failing tests": [report.failing_tests for report in test_runner.reports]
    }, tablefmt="fancy_grid", headers=["Test category", "Number of tests", "Failing tests"]))

    click.echo()
    click.echo(click.style("> Vulnerabilities found by severity", fg="red"))
    click.echo(tabulate({
        "Test category": [report.category.value for report in test_runner.reports],
        "Low": [report.low_severity for report in test_runner.reports],
        "Medium": [report.medium_severity for report in test_runner.reports],
        "High": [report.high_severity for report in test_runner.reports],
    }, tablefmt="fancy_grid", headers=["Test category", "Low", "Medium", "High"]))


if __name__ == "__main__":
    cli()



"""Create product table

Revision ID: 70f9f8166d5f
Revises: 
Create Date: 2023-01-27 14:56:50.279780

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '70f9f8166d5f'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('product',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('price', sa.Float(), nullable=False),
    sa.Column('category', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('product')
    # ### end Alembic commands ###



from pathlib import Path

import pytest
import yaml

from fencer.api_spec import APISpec, BasicEndpoint


@pytest.fixture
def orders_spec():
    return yaml.safe_load((Path(__file__).parent / 'orders_api_spec.yaml').read_text())


def test_load_endpoints(orders_spec):
    api_spec = APISpec(spec=orders_spec, base_url='')
    api_spec.load_endpoints()
    spec_endpoints = [
        endpoint.endpoint for endpoint in api_spec.endpoints
    ]
    expected_endpoints = [
        BasicEndpoint(method='get', path='/orders', base_url=''),
        BasicEndpoint(method='post', path='/orders', base_url=''),
        BasicEndpoint(method='get', path='/orders/{order_id}', base_url=''),
        BasicEndpoint(method='put', path='/orders/{order_id}', base_url=''),
        BasicEndpoint(method='delete', path='/orders/{order_id}', base_url=''),
        BasicEndpoint(method='post', path='/orders/{order_id}/cancel', base_url=''),
    ]
    assert len(expected_endpoints) == len(spec_endpoints)
    for endpoint in expected_endpoints:
        assert endpoint in spec_endpoints


def test_produce_path_parameter():
    pass


def test_capture_query_params():
    pass


def test_produce_query_param_value():
    pass


def test_resolve_component():
    pass


def test_resolve_anyof_component():
    pass


def test_resolve_allof_component():
    pass






import click
import requests

from .api_spec import APISpec
from .test_case import AttackStrategy, TestDescription, HTTPMethods, TestCase, TestResult, \
    VulnerabilitySeverityLevel


class UnauthorizedAccessTestCaseRunner:
    def __init__(self, test_case: TestCase):
        self.test_case = test_case
        self.response = None

    def run(self):
        callable_ = getattr(requests, self.test_case.description.http_method.value.lower())
        self.response = callable_(
            self.test_case.description.url, json=self.test_case.description.payload
        )
        self.resolve_test_result()

    def resolve_test_result(self):
        # If the server fails to respond, endpoint is protected and there's no possibility for exploit,
        # but we can break the server, so we give it a medium severity
        if self.response is None:
            self.test_case.result = TestResult.FAIL
            self.test_case.severity = VulnerabilitySeverityLevel.MEDIUM
            # If response status code is 401 or 403, it's all good
        if self.response.status_code in [401, 403]:
            self.test_case.result = TestResult.SUCCESS
            self.test_case.severity = VulnerabilitySeverityLevel.ZERO
        # If the response status code is in the 2xx status code group, it's pretty bad
        elif self.response.status_code >= 200 < 300:
            self.test_case.result = TestResult.FAIL
            self.test_case.severity = VulnerabilitySeverityLevel.HIGH
        # In all other cases, the response isn't successful, but it's still
        # doing some processing, and that can be leveraged by hackers, so we
        # assign it a high severity
        else:
            self.test_case.result = TestResult.FAIL
            self.test_case.severity = VulnerabilitySeverityLevel.HIGH
        self.test_case.ended_test()


class TestAuthEndpoints:
    def __init__(self, api_spec: APISpec):
        self.api_spec = api_spec
        self.auth_tests = 0

    def test_authorized_endpoints(self):
        failing_tests = []
        for endpoint in self.api_spec.authorized_endpoints:
            click.echo(f"    {endpoint.method.upper()} {endpoint.base_url + endpoint.path.path}", nl=False)
            self.auth_tests += 1
            test_case = UnauthorizedAccessTestCaseRunner(
                test_case=TestCase(
                    category=AttackStrategy.UNAUTHORIZED_ACCESS,
                    test_target="unauthorized_access__access_authorized_endpoints_without_token",
                    description=TestDescription(
                        http_method=getattr(HTTPMethods, endpoint.method.upper()),
                        url=endpoint.safe_url, base_url=endpoint.base_url, path=endpoint.path.path,
                        payload=(
                            endpoint.generate_safe_request_payload()
                            if endpoint.has_request_payload() else None
                        ),
                    )
                )
            )
            test_case.run()
            if test_case.test_case.result == TestResult.FAIL:
                failing_tests.append(test_case.test_case)
                click.echo(" 🚨")
            else:
                click.echo(" ✅")
        return failing_tests



from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Optional

test_case_example = {
    "id": "asdf",
    "timestamp": "2023-02-01:09:00:00TZ:GMT",
    "category": "injection_attack",
    "target_test": "injection_attack__sql_injection__optional_query_parameters",
    "result": "fail",
    "severity": "medium",
    "description": {
        "http_method": "GET",
        "url": "http://localhost:5000/orders?limit=1' 1 OR 1",
        "base_url": "http://localhost:5000",
        "path": "/orders",
        "payload": None,
    }
}


class AttackStrategy(Enum):
    INJECTION = "injection"
    UNAUTHORIZED_ACCESS = "unauthorized_access"
    SURFACE_ATTACKS = "surface_attacks"
    MASS_ASSIGNMENT = "mass_assignment"
    INSECURE_DESIGN = "insecure_design"


class TestResult(Enum):
    SUCCESS = "success"
    FAIL = "fail"
    UNDETERMINED = "undetermined"


class VulnerabilitySeverityLevel(Enum):
    ZERO = "zero"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class HTTPMethods(Enum):
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


@dataclass
class TestDescription:
    http_method: HTTPMethods
    url: str
    base_url: str
    path: str
    payload: Any = None


@dataclass
class TestCase:
    category: AttackStrategy
    test_target: str
    description: TestDescription

    # to be set by the object
    started: Optional[datetime] = None
    ended: Optional[datetime] = None

    # to be set after test run
    result: Optional[TestResult] = None
    severity: Optional[VulnerabilitySeverityLevel] = None

    def __post_init__(self):
        self.started = datetime.now(timezone.utc)

    def ended_test(self):
        self.ended = datetime.now(timezone.utc)

    def dict(self):
        return {
            "started": str(self.started),
            "ended": str(self.ended),
            "category": self.category.value,
            "target_test": self.test_target,
            "result": self.result.value,
            "severity": self.severity.value,
            "description": {
                "http_method": self.description.http_method.value,
                "url": self.description.url,
                "base_url": self.description.base_url,
                "path": self.description.path,
                "payload": self.description.payload,
            }
        }


@dataclass
class TestReporter:
    category: AttackStrategy
    number_tests: int
    failing_tests: int
    low_severity: int = 0
    medium_severity: int = 0
    high_severity: int = 0



import requests

url = "http://localhost:5000/products"


def sql_injection_through_query_param():
    query = "1' OR 1=1 --"
    response = requests.get(url + "?category=" + query)
    print(len(response.json()["products"]))


def check_only_allowed_params():
    response = requests.get(url + "?something=else")
    print(len(response.json()["products"]))


sql_injection_through_query_param()
check_only_allowed_params()



from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    pass


class Product(Base):
    __tablename__ = "product"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    price: Mapped[float]
    category: Mapped[str]



import json
from pathlib import Path

import click

from .api_spec import APISpec
from .authorized_endpoints import TestAuthEndpoints
from .sql_injection import SQLInjectionTestRunner
from .test_case import AttackStrategy, TestCase, VulnerabilitySeverityLevel, TestReporter


class TestRunner:
    def __init__(self, api_spec: APISpec):
        self.api_spec = api_spec
        self.reports = []

    def run_sql_injection_attacks(self):
        sql_injection_test_runner = SQLInjectionTestRunner(api_spec=self.api_spec)

        failing_tests: list[TestCase] = []

        sql_injection_through_query_params_msg = """
  > Testing SQL injection through URL query parameters
          """
        sql_injection_through_path_params_msg = """
  > Testing SQL injection through URL path parameters
          """
        sql_injection_through_request_payloads_msg = """
  > Testing SQL injection through request payloads
          """

        click.echo(sql_injection_through_query_params_msg)
        failing_query_params_tests = sql_injection_test_runner.run_sql_injection_through_query_parameters()

        click.echo(sql_injection_through_path_params_msg)
        failing_path_params_tests = sql_injection_test_runner.run_sql_injection_through_path_parameters()

        click.echo(sql_injection_through_request_payloads_msg)
        failing_payload_tests = sql_injection_test_runner.run_sql_injection_through_request_payloads()

        failing_tests += failing_query_params_tests + failing_path_params_tests + failing_payload_tests

        self.reports.append(TestReporter(
            category=AttackStrategy.INJECTION,
            number_tests=sql_injection_test_runner.injection_tests,
            failing_tests=len(failing_tests),
            low_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.LOW),
            medium_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.MEDIUM),
            high_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.HIGH),
        ))

        failed_tests_file = Path('.fencer/injection_attacks.json')
        failed_tests_file.write_text(
            json.dumps([test.dict() for test in failing_tests], indent=4)
        )

    def run_unauthorized_access_attacks(self):
        test_runner = TestAuthEndpoints(api_spec=self.api_spec)
        failing_tests = test_runner.test_authorized_endpoints()
        self.reports.append(TestReporter(
            category=AttackStrategy.UNAUTHORIZED_ACCESS,
            number_tests=test_runner.auth_tests,
            failing_tests=len(failing_tests),
            low_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.LOW),
            medium_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.MEDIUM),
            high_severity=sum(1 for test in failing_tests if test.severity == VulnerabilitySeverityLevel.HIGH),
        ))
        failed_tests_file = Path('.fencer/unauthorized_access_attacks.json')
        failed_tests_file.write_text(
            json.dumps([test.dict() for test in failing_tests], indent=4)
        )

    def run_surface_attacks(self):
        pass

    def run_mass_assignment_attacks(self):
        pass

    def run_insecure_design_attacks(self):
        pass



import os
import random

from faker import Faker
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from insecure_api.models import Product

db_uri = os.getenv("DB_URI")

assert db_uri is not None, "Need to provide DB_URI environment variable"

session_maker = sessionmaker(bind=create_engine(db_uri))

faker = Faker()

categories = [
    "pens", "cars", "beds", "cups", "lights", "screens", "windows", "shoes"
]

with session_maker() as session:
    products = [
        Product(
            name=faker.word(),
            price=random.randint(1, 100),
            category=random.choice(categories)
        ) for _ in range(500)
    ]

    for product in products:
        session.add(product)

    session.commit()



import random
from dataclasses import dataclass

import click
import requests
from jsf import JSF

from .api_spec import Endpoint, fake_parameter, APISpec
from .test_case import TestResult, TestCase, AttackStrategy, TestDescription, HTTPMethods, VulnerabilitySeverityLevel

sql_injection_strategies = [
    "' OR 1=1 --",
    "' UNION SELECT * FROM information_schema.tables --",
    '"; DROP TABLE users --',
    "'; SELECT user, password FROM users WHERE '1' = '1",
    "'; SELECT id FROM users WHERE '1' = '1",
    "' OR '1' = '1",
    "' OR username LIKE '%",
    ' OR "1"="1"',
    "%' AND 1=0 UNION SELECT * FROM information_schema.tables --",
    "%' OR 1=1; --",
    "' UNION SELECT NULL, table_name FROM information_schema.tables WHERE 2 > 1 \"\"",
]

nosql_injection_strategies = []


@dataclass
class SQLInjectionEndpoint:
    endpoint: Endpoint

    def get_safe_url_path_with_unsafe_required_query_params(self):
        urls = []
        for param in self.endpoint.required_query_params:
            for strategy in sql_injection_strategies:
                param_value = f'?{param["name"]}={strategy}'
                other_params = [
                    other_param for other_param in self.endpoint.required_query_params
                    if other_param['name'] != param['name']
                ]
                if len(other_params) > 0:
                    param_value += '&'
                other_params = '&'.join(
                    f"{other_param['name']}={fake_parameter(param['schema'])}"
                    for other_param in other_params
                )
                url = self.endpoint.safe_url_path_without_query_params + param_value + other_params
                urls.append(url)
        return urls

    def get_safe_url_path_with_unsafe_optional_query_params(self):
        urls = []
        base_url = (
            self.endpoint.safe_url_path_with_safe_required_query_params
            if self.endpoint.has_required_query_params()
            else self.endpoint.safe_url_path_without_query_params
        )
        if self.endpoint.has_required_query_params():
            for param in self.endpoint.optional_query_params:
                for strategy in sql_injection_strategies:
                    param_value = f'?{param["name"]}={strategy}'
                    other_params = [
                        other_param for other_param in self.endpoint.optional_query_params
                        if other_param['name'] != param['name']
                    ]
                    if len(other_params) > 0:
                        param_value += '&'
                    other_params = '&'.join(
                        f"{other_param['name']}={fake_parameter(param['schema'])}"
                        for other_param in other_params
                    )
                    url = base_url + param_value + other_params
                    urls.append(url)
        return urls

    def get_unsafe_url_path_without_query_params(self):
        urls = []
        for param in self.endpoint.path.path_params_list:
            for strategy in sql_injection_strategies:
                path = self.endpoint.path.path.replace(param, strategy)
                urls.append(self.endpoint.base_url + path)
        return urls

    def get_unsafe_url_path_with_safe_required_query_params(self):
        urls = []
        for base_url in self.get_unsafe_url_path_without_query_params():
            urls.append(
                base_url + '?'
                + '&'.join(f"{param['name']}={fake_parameter(param['schema'])}"
                           for param in self.endpoint.required_query_params)
            )
        return urls

    def get_urls_with_unsafe_query_params(self):
        urls = []
        if self.endpoint.has_required_query_params():
            urls.extend(self.get_safe_url_path_with_unsafe_required_query_params())
        if self.endpoint.has_optional_query_params():
            urls.extend(self.get_safe_url_path_with_unsafe_optional_query_params())
        for url in urls:
            yield url

    def get_urls_with_unsafe_path_params(self):
        urls = []
        if self.endpoint.path.has_path_params():
            urls.extend(self.get_unsafe_url_path_without_query_params())
            if self.endpoint.has_required_query_params():
                urls.extend(self.get_unsafe_url_path_with_safe_required_query_params())
        for url in urls:
            yield url

    def _inject_dangerous_sql_in_payload(self, payload, schema):
        # need to include anyOf, allOf
        if schema['type'] == 'array':
            return [
                self._inject_dangerous_sql_in_payload(item, schema['items'])
                for item in payload
            ]
        if schema['type'] == 'object':
            # sometimes properties aren't specified so soft access
            for name, description in schema.get('properties', {}).items():
                # property may not be required
                if name not in payload:
                    continue
                if description['type'] == 'string':
                    payload[name] = random.choice(sql_injection_strategies)
                if description['type'] == 'array':
                    payload[name] = self._inject_dangerous_sql_in_payload(
                        payload[name], description
                    )
        return payload

    def generate_unsafe_request_payload(self):
        # this should be plural returning an array of payloads with different
        # sql injection strategies
        schema = self.endpoint.body['content']['application/json']['schema']
        if 'allOf' in schema:
            schema = schema['allOf'][0]
        if 'anyOf' in schema:
            schema = schema['anyOf'][0]
        payload = JSF(schema).generate()
        return self._inject_dangerous_sql_in_payload(payload, schema)


class InjectionTestCaseRunner:
    def __init__(self, test_case: TestCase):
        self.test_case = test_case
        self.response = None

    def run(self):
        callable_ = getattr(requests, self.test_case.description.http_method.value.lower())
        self.response = callable_(
            self.test_case.description.url, json=self.test_case.description.payload
        )
        self.resolve_test_result()

    def resolve_test_result(self):
        """
        In this case, it's difficult to assess the severity of the failure without looking
        at the backend logs. We'll assume that:
        - Failure to response indicates major outage caused by the request
        - 500 status code indicates potential high severity and potential for leaking traceback
        Everything else is severity Zero.
        Until we can develop better heuristics for response analysis, this is the best we can do.
        """
        # If the server fails to respond, we assume we broke it
        if self.response is None:
            self.test_case.result = TestResult.FAIL
            self.test_case.severity = VulnerabilitySeverityLevel.HIGH
        # If the request causes a server error, it likely broke it
        elif self.response.status_code >= 500:
            self.test_case.result = TestResult.FAIL
            self.test_case.severity = VulnerabilitySeverityLevel.HIGH
        # Any status code below 500 indicates the response was correctly processed
        # (i.e. correctly accepted or rejected)
        else:
            self.test_case.result = TestResult.SUCCESS
            self.test_case.severity = VulnerabilitySeverityLevel.ZERO
        self.test_case.ended_test()


class SQLInjectionTestRunner:
    def __init__(self, api_spec: APISpec):
        self.api_spec = api_spec
        self.injection_tests = 0
        self.reports = []

    def run_sql_injection_through_query_parameters(self):
        failing_tests = []
        for endpoint in self.api_spec.endpoints:
            sql_injection = SQLInjectionEndpoint(endpoint)
            endpoint_failing_tests = []
            click.echo(f"    {endpoint.method.upper()} {endpoint.base_url + endpoint.path.path}", nl=False)
            for url in sql_injection.get_urls_with_unsafe_query_params():
                self.injection_tests += 1
                test_case = InjectionTestCaseRunner(
                    test_case=TestCase(
                        category=AttackStrategy.INJECTION,
                        test_target="sql_injection__optional_query_parameters",
                        description=TestDescription(
                            http_method=getattr(HTTPMethods, endpoint.method.upper()),
                            url=url, base_url=endpoint.base_url, path=endpoint.path.path,
                            payload=endpoint.generate_safe_request_payload() if endpoint.has_request_payload() else None,
                        )
                    )
                )
                test_case.run()
                if test_case.test_case.result == TestResult.FAIL:
                    endpoint_failing_tests.append(test_case.test_case)
            if len(endpoint_failing_tests) > 0:
                failing_tests.extend(endpoint_failing_tests)
                click.echo(" 🚨")
            else:
                click.echo(" ✅")
        return failing_tests

    def run_sql_injection_through_path_parameters(self):
        failing_tests = []
        for endpoint in self.api_spec.endpoints:
            if not endpoint.has_path_params():
                continue
            sql_injection = SQLInjectionEndpoint(endpoint)
            endpoint_failing_tests = []
            click.echo(f"    {endpoint.method.upper()} {endpoint.base_url + endpoint.path.path}", nl=False)
            for url in sql_injection.get_urls_with_unsafe_path_params():
                self.injection_tests += 1
                test_case = InjectionTestCaseRunner(
                    test_case=TestCase(
                        category=AttackStrategy.INJECTION,
                        test_target="sql_injection__optional_query_parameters",
                        description=TestDescription(
                            http_method=getattr(HTTPMethods, endpoint.method.upper()),
                            url=url, base_url=endpoint.base_url, path=endpoint.path.path,
                            payload=endpoint.generate_safe_request_payload() if endpoint.has_request_payload() else None,
                        )
                    )
                )
                test_case.run()
                if test_case.test_case.result == TestResult.FAIL:
                    endpoint_failing_tests.append(test_case.test_case)
            if len(endpoint_failing_tests) > 0:
                failing_tests.extend(endpoint_failing_tests)
                click.echo(" 🚨")
            else:
                click.echo(" ✅")
        return failing_tests

    def run_sql_injection_through_request_payloads(self):
        failing_tests = []
        for endpoint in self.api_spec.endpoints:
            if not endpoint.has_request_payload():
                continue
            sql_injection = SQLInjectionEndpoint(endpoint)
            click.echo(f"    {endpoint.method.upper()} {endpoint.base_url + endpoint.path.path}", nl=False)
            self.injection_tests += 1
            test_case = InjectionTestCaseRunner(
                test_case=TestCase(
                    category=AttackStrategy.INJECTION,
                    test_target="sql_injection__optional_query_parameters",
                    description=TestDescription(
                        http_method=getattr(HTTPMethods, endpoint.method.upper()),
                        url=endpoint.safe_url, base_url=endpoint.base_url, path=endpoint.path.path,
                        payload=sql_injection.generate_unsafe_request_payload()
                    )
                )
            )
            test_case.run()
            if test_case.test_case.result == TestResult.FAIL:
                failing_tests.append(test_case.test_case)
                click.echo(" 🚨")
            else:
                click.echo(" ✅")
        return failing_tests



import random
import re
import string
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Union, Optional

import exrex
from jsf import JSF

standard_http_methods = ['get', 'post', 'put', 'patch', 'delete', 'options', 'head']


@dataclass
class NumberRanges:
    schema: dict

    @property
    def minimum(self):
        minimum = self.schema.get('minimum', 0)
        if self.schema.get('exclusiveMinimum'):
            minimum = self.schema['exclusiveMinimum'] + 1
        return minimum

    @property
    def maximum(self):
        total_max = 2147483647 if self.schema.get('format', '') == 'int32' else 9223372036854775807
        maximum = self.schema.get('maximum', total_max)
        if self.schema.get('exclusiveMaximum'):
            maximum = self.schema['exclusiveMaximum'] - 1
        return maximum


def fake_parameter(schema):
    if schema.get('example'):
        return schema['example']
    if schema.get('default'):
        return schema['default']
    if schema['type'] == 'string':
        if 'format' not in schema:
            if schema.get('pattern'):
                return exrex.getone(schema['pattern'])
            return ''.join(
                random.choice(string.ascii_letters)
                for _ in range(schema.get('minLength', 2), schema.get('maxLength', 20))
            )
        if schema['format'] == 'uuid':
            return str(uuid.uuid4())
        if schema['format'] == 'date':
            return str(datetime.now().date())
        if schema['format'] == 'date-time':
            return str(datetime.now())
        if schema['format'] == 'email':
            return 'test@example.com'
        if schema['format'] == 'ipv4':
            return '127.0.0.1'

    if schema['type'] == 'integer':
        ranges = NumberRanges(schema)
        return random.randint(ranges.minimum, ranges.maximum)
    if schema['type'] == 'number':
        if 'format' not in schema:
            ranges = NumberRanges(schema)
            return random.randint(ranges.minimum, ranges.maximum)
        if schema['format'] == 'float':
            return random.random() * 1000
        if schema['format'] == 'double':
            return round(random.random() * 1000, 2)

    if schema['type'] == 'boolean':
        return random.choice(['true', 'false'])


@dataclass
class BasicEndpoint:
    method: str
    path: str
    base_url: str

    def __str__(self):
        return f"{self.method.upper()} {self.base_url}{self.base_url}"


@dataclass
class Endpoint:
    base_url: str
    api_path: str
    method: str
    parameters: list
    body: Optional[dict]
    responses: dict
    security: Optional[Union[dict, list]] = None

    def __post_init__(self):
        self.path = APIPath(
            path=self.api_path,
            path_params_schemas=self.path_params
        )
        self.endpoint = BasicEndpoint(
            method=self.method, path=self.api_path, base_url=self.base_url
        )

    def __str__(self):
        return str(self.endpoint)

    @property
    def query_params(self):
        return [
            param for param in self.parameters if param['in'] == 'query'
        ]

    @property
    def required_query_params(self):
        return [
            param for param in self.query_params if param['required']
        ]

    @property
    def optional_query_params(self):
        return [
            param for param in self.parameters if not param['required']
        ]

    @property
    def path_params(self):
        return [
            param for param in self.parameters if param['in'] == 'path'
        ]

    def has_query_params(self):
        return len(self.query_params) > 0

    def has_required_query_params(self):
        return len(self.required_query_params) > 0

    def has_optional_query_params(self):
        return len(self.optional_query_params) > 0

    def has_path_params(self):
        return len(self.path_params) > 0

    @property
    def safe_url_path_without_query_params(self):
        return self.base_url + self.path.build_safe_path()

    @property
    def safe_url_path_with_safe_required_query_params(self):
        return (
                self.safe_url_path_without_query_params + '?'
                + '&'.join(f"{param['name']}={fake_parameter(param['schema'])}"
                           for param in self.required_query_params)
        )

    @property
    def safe_url(self):
        if self.has_required_query_params():
            return self.safe_url_path_with_safe_required_query_params
        return self.safe_url_path_without_query_params

    def has_request_payload(self):
        if self.body is None:
            return False
        return self.body.get('content', {}).get('application/json', {}).get('schema') is not None

    def generate_safe_request_payload(self):
        schema = self.body['content']['application/json']['schema']
        return JSF(schema).generate()


@dataclass
class APIPath:
    path: str
    path_params_schemas: list

    path_param_regex = re.compile('{.*?}')

    def __post_init__(self):
        self.path_params_list = self.path_param_regex.findall(self.path)
        self._undocumented_path_params = [
            param for param in self.path_params_schemas if param['name'] not in self.path
        ]

    def has_path_params(self):
        return len(self.path_params_list) > 0

    def has_undocumented_path_params(self):
        return len(self._undocumented_path_params) > 0

    def build_safe_path(self):
        if not self.has_path_params():
            return self.path

        path = self.path
        for param in self.path_params_schemas:
            path = path.replace(
                f'{{{param["name"]}}}',
                fake_parameter(param['schema']),
            )

        if not self.has_undocumented_path_params():
            return path

        for param in self._undocumented_path_params:
            path = path.replace(
                f'{{{param["name"]}}}',
                fake_parameter({'type': 'string'}),
            )

        return path


class APISpec:

    def __init__(self, base_url: str, spec: dict):
        self.base_url = base_url
        self.spec = spec
        self.servers = spec.get('servers')
        self.paths = spec['paths']
        self.components = spec['components']
        self.endpoints: list[Endpoint] = []

    @property
    def authorized_endpoints(self):
        if "securitySchemes" not in self.components:
            return []
        # If the spec doesn't have a global security requirement,
        # we filter for endpoints with a declared security scheme
        if "security" not in self.spec:
            return [
                endpoint for endpoint in self.endpoints
                if endpoint.security is not None and len(endpoint.security) > 0
            ]
        # If the spec has a global security requirement, we filter out
        # endpoints that override the global requirement with an empty
        # object or an empty array
        return [
            endpoint for endpoint in self.endpoints
            if endpoint.security is None or len(endpoint.security) > 0
        ]

    def load_endpoints(self):
        paths = self.paths.keys()
        for path in paths:
            http_methods = [method for method in self.paths[path].keys() if method in standard_http_methods]
            url_params = self.paths[path].get('parameters', [])
            # do options request and confirm only these methods are allowed
            for method in http_methods:
                self.endpoints.append(
                    Endpoint(
                        base_url=self.base_url,
                        api_path=path,
                        method=method,
                        parameters=url_params + self.paths[path][method].get('parameters', []),
                        body=self.resolve_body(self.paths[path][method].get('requestBody')),
                        responses=self.paths[path][method].get('responses'),
                        security=self.paths[path][method].get('security'),
                    )
                )

    def resolve_body(self, body: Optional[dict]):
        if body is None:
            return

        if body.get('content') is None:
            return

        # we only support application/json atm
        if body['content'].get('application/json') is None:
            return

        if '$ref' in body['content']['application/json']['schema']:
            schema = self.resolve_schema(body['content']['application/json']['schema']['$ref'])
            body['content']['application/json']['schema'] = schema

        if 'allOf' in body['content']['application/json']['schema']:
            for index, schema in enumerate(body['content']['application/json']['schema']['allOf']):
                if '$ref' in schema:
                    schema = self.resolve_schema(schema['$ref'])
                    body['content']['application/json']['schema']['allOf'][index] = schema

        if 'anyOf' in body['content']['application/json']['schema']:
            for index, schema in enumerate(body['content']['application/json']['schema']['anyOf']):
                if '$ref' in schema:
                    schema = self.resolve_schema(schema['$ref'])
                    body['content']['application/json']['schema']['anyOf'][index] = schema

        return body

    def resolve_schema(self, schema_ref):
        schema_name = schema_ref.split('/')[-1]
        schema = self.spec['components']['schemas'][schema_name]

        if 'allOf' in schema:
            raise Exception('allOf not implemented')

        if 'anyOf' in schema:
            raise Exception('anyOf not implemented')

        for name, description in schema['properties'].items():
            if '$ref' in description:
                property_schema = self.resolve_schema(description['$ref'])
                schema['properties'][name] = property_schema
                continue

            if description['type'] == 'array' and '$ref' in description['items']:
                items_schema = self.resolve_schema(description['items']['$ref'])
                description['items'] = items_schema

        return schema
